﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace IcapClientDll
{
	public class IcapConnection: IDisposable
	{
		private string m_strHostName;
		public string HostName
		{
			get => m_strHostName;
			set => m_strHostName = value;
		}

		private int m_iPort;
		public int Port
		{
			get => m_iPort;
			set => m_iPort = value;
		}

		private string m_strServiceName;
		public string ServiceName
		{
			get => m_strServiceName;
			set => m_strServiceName = value;
		}

		private string m_strMethod;
		public string Method
		{
			get => m_strMethod;
			set => m_strMethod = value;
		}

		private bool m_bIsconnected;
		public bool IsConnected
		{
			get => m_bIsconnected;
			set => m_bIsconnected = value;
		}

		private string m_strError;
		public string Error
		{
			get => m_strError;
			set => m_strError = value;
		}

		private Socket sender;

		private int iPreviewSize;

		public string strOptionsRequestPayload;

		private IcapConnection()
		{
			m_strError = string.Empty;
		}

		public IcapConnection(string m_strHostName, int m_iPort) : this()
		{
			this.m_strHostName = m_strHostName;
			this.m_iPort = m_iPort;

			fnConnect();
		}

		public IcapConnection(string m_strHostName, int m_iPort, string m_strServiceName, string m_strMethod, int iPreviewSize = -1) : this(m_strHostName, m_iPort)
		{
			this.m_strServiceName = m_strServiceName;
			this.m_strMethod = m_strMethod;

			if (iPreviewSize == -1)
			{
				fnGetPreviewSizeFromServer();
			}
			else
			{
				this.iPreviewSize = iPreviewSize;
			}
		}

		private void fnConnect()
		{
			try
			{
				/* Initialize connection */
				IPAddress ipAddress = IPAddress.Parse(m_strHostName);
				IPEndPoint remoteEP = new IPEndPoint(ipAddress, m_iPort);

				/* Create a TCP/IP  socket */
				sender = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
				sender.Connect(remoteEP);

				m_bIsconnected = true;
			}
			catch (Exception ex)
			{
				m_bIsconnected = false;
				m_strError = ex.Message;
			}
		}

		private void fnGetPreviewSizeFromServer()
		{
			try
			{
				Dictionary<string, string> responseMap = fnParseHeader(fnGetServerOptions());

				responseMap.TryGetValue("StatusCode", out string tempString);
				if (tempString != null)
				{
					int status = Convert.ToInt16(tempString);

					switch (status)
					{
						case 200:
							responseMap.TryGetValue("Preview", out tempString);
							if (tempString != null)
							{
								iPreviewSize = Convert.ToInt16(tempString);
							}; break;
						default: throw new Exception("Could not get preview size from server");
					}
				}
				else
				{
					throw new Exception("Could not get options from server");
				}
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
			}
		}

		public string fnGetServerOptions()
		{
			try
			{
				strOptionsRequestPayload =
				"OPTIONS icap://" + m_strHostName + ":" + m_iPort.ToString() + "/" + m_strServiceName + " ICAP/" + IcapConstants.VERSION + "\r\n"
				+ "Host: " + m_strHostName + ":" + m_iPort.ToString() + "\r\n"
				+ "User-Agent: " + IcapConstants.USERAGENT + "\r\n"
				+ "Encapsulated: null-body=0\r\n"
				+ "\r\n";

				byte [] msg = Encoding.ASCII.GetBytes(strOptionsRequestPayload);
				sender.Send(msg);

				return fnGetHeader(IcapConstants.ICAPTERMINATOR);
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
				return string.Empty;
			}
		}

		/// <summary>
		/// Given a raw response header as a String, it will parse through it and return a Dictionary of the result
		/// </summary>
		/// <param name="response">A raw response header as a String.</parm>
		/// <returns>Dictionary of the key,value pairs of the response</returns>
		private Dictionary<string, string> fnParseHeader(string response)
		{
			Dictionary<string, string> headers = new Dictionary<string, string>();

			/****SAMPLE:****
             * ICAP/1.0 204 Unmodified
             * Server: C-ICAP/0.1.6
             * Connection: keep-alive
             * ISTag: CI0001-000-0978-6918203
             */

			/* The status code is located between the first 2 whitespaces */
			/* Read status code */
			int x = response.IndexOf(" ", 0);
			int y = response.IndexOf(" ", x + 1);
			string statusCode = response.Substring(x + 1, y - x - 1);
			headers.Add("StatusCode", statusCode);

			/*
             * Each line in the sample is ended with "\r\n". 
             * When (i+2==response.length()) The end of the header have been reached.
             * The +=2 is added to skip the "\r\n".
             * Read headers
             */
			int i = response.IndexOf("\r\n", y);
			i += 2;
			while (i + 2 != response.Length && response.Substring(i).Contains(":"))
			{
				int n = response.IndexOf(":", i);
				string key = response.Substring(i, n - i);

				n += 2;
				i = response.IndexOf("\r\n", n);
				string value = response.Substring(n, i - n);

				headers.Add(key, value);
				i += 2;
			}
			return headers;
		}

		/// <summary>
		/// Receive an expected ICAP header as response of a request. The returned String should be parsed with fnParseHeader().
		/// </summary>
		/// <param name="terminator">Relative or absolute filepath to a file.</parm>
		/// <exception cref="Exception">Thrown when error occurs in communication with server</exception>
		/// <returns>String of the raw response.</returns>
		private string fnGetHeader(string terminator)
		{
			try
			{
				/* Initialisations */
				byte [] endofheader = Encoding.UTF8.GetBytes(terminator);
				byte [] buffer = new byte [IcapConstants.ReceiveLength];

				int n;
				int offset = 0;

				/* stdRecieveLength-offset is replaced by '1' to not receive the next (HTTP) header */

				while ((offset < IcapConstants.ReceiveLength) && ((n = sender.Receive(buffer, offset, 1, SocketFlags.None)) != 0)) // first part is to secure against DOS
				{
					offset += n;
					if (offset > endofheader.Length + 13)
					/* 13 is the smallest possible message (ICAP/1.0 xxx\r\n) or (HTTP/1.0 xxx\r\n) */
					{
						byte [] lastBytes = new byte [endofheader.Length];
						Array.Copy(buffer, offset - endofheader.Length, lastBytes, 0, endofheader.Length);
						if (endofheader.SequenceEqual(lastBytes))
						{
							return Encoding.ASCII.GetString(buffer, 0, offset);
						}
					}
				}

				throw new Exception("Error in getHeader() method");
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
				return string.Empty;
			}
		}

		public List<IcapResponse> fnScanFiles_RESPMOD(List<string> lstFilesToScan)
		{
			/* SAMPLE
			 * C: RESPMOD icap://172.16.11.19:1344/SYMCScanResp-AV-DDR ICAP/1.0
				C: Host: 172.16.11.19:1344
				C: Connection: close
				C: Encapsulated: req-hdr=0, res-hdr=52, res-body=71
				C:
				C: get eicar.com HTTP/1.1
				C: Host:icheck.symantec.com
				C:
				C: HTTP/1.1 200 OK
				C:
				C: 44
				C: Sending chunk of size 68 bytes
				C:
				C: 0
				C:
				*/

			try
			{
				/* Global Initialisations */
				List<IcapResponse> result = new List<IcapResponse>();
				string tempString = string.Empty;

				if (lstFilesToScan != null && lstFilesToScan.Count > 0)
				{
					foreach (string strFile in lstFilesToScan)
					{
						if (File.Exists(strFile) == true)
						{
							IcapResponse icapResponse = new IcapResponse
							{
								File = strFile
							};

							using (FileStream fileStream = new FileStream(strFile, FileMode.Open))
							{
								int fileSize = (int)fileStream.Length;

								//First part of header
								string resBody = "Content-Length: " + fileSize + "\r\n\r\n";

								int previewSize = iPreviewSize;
								if (fileSize < iPreviewSize)
								{
									previewSize = fileSize;
								}

								string strPayload =
									"RESPMOD icap://" + m_strHostName + ":" + m_iPort.ToString() + "/" + m_strServiceName + " ICAP/" + IcapConstants.VERSION + "\r\n"
									+ "Host: " + m_strHostName + ":" + m_iPort.ToString() + "\r\n"
									+ "User-Agent: " + IcapConstants.USERAGENT + "\r\n"
									+ "Allow: 204\r\n"
									+ "Preview: " + previewSize + "\r\n"
									+ "Encapsulated: res-hdr=0, res-body=" + resBody.Length + "\r\n"
									+ "\r\n"
									+ resBody
									+ previewSize.ToString("X") + "\r\n";
								byte [] requestBuffer = Encoding.ASCII.GetBytes(strPayload);

								icapResponse.RequestPaylod = strPayload;
								sender.Send(requestBuffer);

								byte [] chunk = new byte [previewSize];

								fileStream.Read(chunk, 0, previewSize);
								sender.Send(chunk);
								sender.Send(Encoding.ASCII.GetBytes("\r\n"));
								if (fileSize <= previewSize)
								{
									sender.Send(Encoding.ASCII.GetBytes("0; ieof\r\n\r\n"));
								}
								else if (previewSize != 0)
								{
									sender.Send(Encoding.ASCII.GetBytes("0\r\n\r\n"));
								}

								// Parse the response! It might not be "100 continue".
								// if fileSize<previewSize, then the stream waiting is actually the allowed/disallowed signal
								// otherwise it is a "go" for the rest of the file.
								Dictionary<string, string> responseMap = new Dictionary<string, string>();
								int status;

								if (fileSize > previewSize)
								{
									string parseMe = fnGetHeader(IcapConstants.ICAPTERMINATOR);
									responseMap = fnParseHeader(parseMe);

									icapResponse.Response = parseMe;

									responseMap.TryGetValue("StatusCode", out tempString);
									if (tempString != null)
									{
										status = Convert.ToInt16(tempString);

										switch (status)
										{
											case 100: break;
											case 200: icapResponse.Status = false; break;
											case 204: icapResponse.Status = true; break;
											case 404:
												result.Add(icapResponse);
												throw new Exception("404: ICAP Service not found");
											default:
												result.Add(icapResponse);
												throw new Exception("Server returned unknown status code:" + status);
										}
									}
								}

								//Sending remaining part of file
								if (fileSize > previewSize)
								{
									int offset = previewSize;
									int n;
									byte [] buffer = new byte [IcapConstants.SendLength];
									while ((n = fileStream.Read(buffer, 0, IcapConstants.SendLength)) > 0)
									{
										offset += n;  // offset for next reading
										sender.Send(Encoding.ASCII.GetBytes(buffer.Length.ToString("X") + "\r\n"));
										sender.Send(buffer);
										sender.Send(Encoding.ASCII.GetBytes("\r\n"));
									}
									//Closing file transfer.
									sender.Send(Encoding.ASCII.GetBytes("0\r\n\r\n"));
								}
								//fileStream.Close();

								responseMap.Clear();
								string response = fnGetHeader(IcapConstants.ICAPTERMINATOR);
								responseMap = fnParseHeader(response);

								icapResponse.Response = response;

								responseMap.TryGetValue("StatusCode", out tempString);
								if (tempString != null)
								{
									status = Convert.ToInt16(tempString);

									if (status == 204) { icapResponse.Status = true; } //Unmodified

									if (status == 200) //OK - The ICAP status is ok, but the encapsulated HTTP status will likely be different
									{
										response = fnGetHeader(IcapConstants.HTTPTERMINATOR);
										// Searching for: <title>ProxyAV: Access Denied</title>
										try
										{
											int x = response.IndexOf("<title>", 0);
											int y = response.IndexOf("</title>", x);
											string statusCode = response.Substring(x + 7, y - x - 7);

											if (statusCode.Equals("ProxyAV: Access Denied"))
											{
												icapResponse.Status = false;
											}
										}
										catch (Exception) { }
									}
								}
								else
								{
									throw new Exception("Unrecognized or no status code in response header.");
								}
								result.Add(icapResponse);
							}
						}
					}
				}

				return result;
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
				return null;
			}
			finally
			{
			}
		}

		public List<IcapResponse> fnScanFiles_REQMOD(List<string> lstFilesToScan)
		{
			/* SAMPLE
			 * C: REQMOD icap://172.16.11.19:1344/SYMCScanReq-AV-URL ICAP/1.0
				C: Host: 172.16.11.19:1344
				C: Connection: close
				C: Encapsulated: req-hdr=0, req-body=147
				C:
				C: POST annacar-lzh-rar-hqx-exe.uue HTTP/1.1
				C: Host:icheck.symantec.com
				C: Accept: text/html, text/plain
				C: Accept-Encoding: compress
				C: Pragma: no-cache
				C:
				C: 44
				C: Sending chunk of size 68 bytes
				C:
				C: 0
				C:
			*/

			try
			{
				/* Global Initialisations */
				List<IcapResponse> result = new List<IcapResponse>();
				string tempString = string.Empty;

				if (lstFilesToScan != null && lstFilesToScan.Count > 0)
				{
					foreach (string strFile in lstFilesToScan)
					{
						if (File.Exists(strFile) == true)
						{
							IcapResponse icapResponse = new IcapResponse
							{
								File = strFile
							};

							using (FileStream fileStream = new FileStream(strFile, FileMode.Open))
							{
								int fileSize = (int)fileStream.Length;

								//First part of header
								string resBody = "Content-Length: " + fileSize + "\r\n\r\n";

								int previewSize = iPreviewSize;
								if (fileSize < iPreviewSize)
								{
									previewSize = fileSize;
								}

								string strPayload =
									"REQMOD icap://" + m_strHostName + ":" + m_iPort.ToString() + "/" + m_strServiceName + " ICAP/" + IcapConstants.VERSION + "\r\n"
									+ "Host: " + m_strHostName + ":" + m_iPort.ToString() + "\r\n"
									+ "User-Agent: " + IcapConstants.USERAGENT + "\r\n"
									+ "Connection: close\r\n"
									+ "Preview: " + previewSize + "\r\n"
									+ "Encapsulated: req-hdr=0, req-body=" + resBody.Length + "\r\n"
									+ "\r\n"
									+ "POST annacar-lzh-rar-hqx-exe.uue HTTP/1.1" + "\r\n"
									+ "Host: icheck.symantec.com" + "\r\n"
									+ "Accept: text/html, text/plain" + "\r\n"
									+ "Accept-Encoding: compress" + "\r\n"
									+ "Pragma: no-cache" + "\r\n"
									+ "\r\n"
									+ "44" + "\r\n"
									+ "Sending chunk of size 68 bytes" + "\r\n"
									+ "\r\n"
									+ "0";
								byte [] requestBuffer = Encoding.ASCII.GetBytes(strPayload);

								icapResponse.RequestPaylod = strPayload;
								sender.Send(requestBuffer);

								byte [] chunk = new byte [previewSize];

								fileStream.Read(chunk, 0, previewSize);
								sender.Send(chunk);
								sender.Send(Encoding.ASCII.GetBytes("\r\n"));
								if (fileSize <= previewSize)
								{
									sender.Send(Encoding.ASCII.GetBytes("0; ieof\r\n\r\n"));
								}
								else if (previewSize != 0)
								{
									sender.Send(Encoding.ASCII.GetBytes("0\r\n\r\n"));
								}

								// Parse the response! It might not be "100 continue".
								// if fileSize<previewSize, then the stream waiting is actually the allowed/disallowed signal
								// otherwise it is a "go" for the rest of the file.
								Dictionary<string, string> responseMap = new Dictionary<string, string>();
								int status;

								if (fileSize > previewSize)
								{
									string parseMe = fnGetHeader(IcapConstants.ICAPTERMINATOR);
									responseMap = fnParseHeader(parseMe);

									icapResponse.Response = parseMe;

									responseMap.TryGetValue("StatusCode", out tempString);
									if (tempString != null)
									{
										status = Convert.ToInt16(tempString);

										switch (status)
										{
											case 100: break;
											case 200: icapResponse.Status = false; break;
											case 204: icapResponse.Status = true; break;
											case 404:
												result.Add(icapResponse);
												throw new Exception("404: ICAP Service not found");
											default:
												result.Add(icapResponse);
												throw new Exception("Server returned unknown status code:" + status);
										}
									}
								}

								//Sending remaining part of file
								if (fileSize > previewSize)
								{
									int offset = previewSize;
									int n;
									byte [] buffer = new byte [IcapConstants.SendLength];
									while ((n = fileStream.Read(buffer, 0, IcapConstants.SendLength)) > 0)
									{
										offset += n;  // offset for next reading
										sender.Send(Encoding.ASCII.GetBytes(buffer.Length.ToString("X") + "\r\n"));
										sender.Send(buffer);
										sender.Send(Encoding.ASCII.GetBytes("\r\n"));
									}
									//Closing file transfer.
									sender.Send(Encoding.ASCII.GetBytes("0\r\n\r\n"));
								}
								//fileStream.Close();

								responseMap.Clear();
								string response = fnGetHeader(IcapConstants.ICAPTERMINATOR);
								responseMap = fnParseHeader(response);

								icapResponse.Response = response;

								responseMap.TryGetValue("StatusCode", out tempString);
								if (tempString != null)
								{
									status = Convert.ToInt16(tempString);

									if (status == 204) { icapResponse.Status = true; } //Unmodified

									if (status == 200) //OK - The ICAP status is ok, but the encapsulated HTTP status will likely be different
									{
										response = fnGetHeader(IcapConstants.HTTPTERMINATOR);
										// Searching for: <title>ProxyAV: Access Denied</title>
										try
										{
											int x = response.IndexOf("<title>", 0);
											int y = response.IndexOf("</title>", x);
											string statusCode = response.Substring(x + 7, y - x - 7);

											if (statusCode.Equals("ProxyAV: Access Denied"))
											{
												icapResponse.Status = false;
											}
										}
										catch (Exception) { }
									}
								}
								else
								{
									throw new Exception("Unrecognized or no status code in response header.");
								}
								result.Add(icapResponse);
							}
						}
					}
				}

				return result;
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
				return null;
			}
			finally
			{
			}
		}

		public List<IcapResponse> fnScanFiles_FILEMOD(List<string> lstFilesToScan)
		{
			/* SAMPLE
			 * C: FILEMOD icap://172.16.11.19:1344/SYMCScanResp-AV ICAP/1.0
				C: Host: 172.16.11.19:1344
				C: X-Filepath: d:\vfiles\test2\frere.exe
				C: Connection: close
				C: Encapsulated: null-body=0
				C:
				*/
			try
			{
				/* Global Initialisations */
				List<IcapResponse> result = new List<IcapResponse>();
				string tempString = string.Empty;

				if (lstFilesToScan != null && lstFilesToScan.Count > 0)
				{
					foreach (string strFile in lstFilesToScan)
					{
						if (File.Exists(strFile) == true)
						{
							IcapResponse icapResponse = new IcapResponse
							{
								File = strFile
							};

							string strPayload = string.Format("FILEMOD icap://{0}:{1}/{2} ICAP/{3}{4}Host: {0}:{1}{4}X-Filepath: {5}{4}Connection: close{4}Encapsulated: null-body=0", m_strHostName, m_iPort.ToString(), m_strServiceName, IcapConstants.VERSION, "\r\n", strFile);

							icapResponse.RequestPaylod = strPayload;
							byte [] requestBuffer = Encoding.ASCII.GetBytes(strPayload);

							sender.Send(requestBuffer);
							Dictionary<string, string> responseMap = new Dictionary<string, string>();

							string parseMe = fnGetHeader(IcapConstants.ICAPTERMINATOR);
							responseMap = fnParseHeader(parseMe);

							icapResponse.Response = parseMe;

							responseMap.TryGetValue("StatusCode", out tempString);
							if (tempString != null)
							{
								short status = Convert.ToInt16(tempString);
							}
							else
							{
								throw new Exception("Unrecognized or no status code in response header.");
							}

							result.Add(icapResponse);
						}
					}
				}

				return result;
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
				return null;
			}
			finally
			{
			}
		}

		public void Dispose()
		{
			try
			{
				sender.Shutdown(SocketShutdown.Both);
				sender.Close();
				sender.Dispose();
			}
			catch (Exception ex)
			{
				m_strError = ex.Message;
			}
		}
	}
}
